# Author: Guillaume Roullet <roullet@univ-brest.fr>

"""ncpymovie: generate a video file from a list of netcdf

this is a standalone script that can be called from the shell

the user needs to provide

    - the variable to plot
    - the netcdf file or the list of them (wildcards ok)
    - the level on which to plot (for 3D variables)

optional arguments include

    - the bounds of the color axis
    - the outfile

"""
from __future__ import print_function
import matplotlib

matplotlib.use('TkAgg')
font = {'size': 16}
matplotlib.rc('font', **font)


from shutil import which
import argparse
import glob
import time
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as ani
from mpl_toolkits.axes_grid1 import make_axes_locatable
import ncpy

plt.ion()


def get_pixel_size_of_axis(ax):
    """return the (width, height) of the axis 'ax' in pixels"""

    bounds = ax.get_position().bounds
    fig_dpi = ax.get_figure().dpi
    fig_siz = ax.get_figure().get_size_inches()
    return [bounds[2]*fig_siz[0]*fig_dpi, bounds[3]*fig_siz[1]*fig_dpi]


def init_figure(fig, time, field2d):
    """create the figure

    this is where you adapt the figure to your needs
    the function should return the graphical objects to update
    """

    field_size = np.shape(field2d)
    zoom_x = (0.8*fig_size[0])//field_size[1]
    zoom_y = (0.9*fig_size[1])//field_size[0]
    # print("zoom factors x : %i / y : %i" % (zoom_x, zoom_y))
    zoom_factor = min(zoom_x, zoom_y)
    print("each grid cell is (%i, %i) pixels" % (zoom_factor, zoom_factor))
    rectangle = [0.1, 0.1,
                 field_size[1]/fig_size[0]*zoom_factor,
                 field_size[0]/fig_size[1]*zoom_factor]
    ax = plt.axes(rectangle)

    im = ax.imshow(field2d, cmap=plt.get_cmap('RdBu_r'),
                   vmin=cax[0], vmax=cax[1], origin='lower')
    ti = ax.set_title((title_string % varname)+time_into_string(time))
    ax.set_xlabel('X')
    ax.set_ylabel('Y')

    divider = make_axes_locatable(ax)
    cbax = divider.append_axes("right", size="3%", pad=0.1)

#    pos = [rectangle[0]+rectangle[2]+0.02, rectangle[1], 0.05, rectangle[3]]
    cb = fig.colorbar(im, cax=cbax)

    pos = np.array(cb.ax.get_position().bounds)
    pos[1], pos[3] = rectangle[1], rectangle[3]
#    cb.ax.set_position(pos)
#    fig.canvas.draw()
    return im, ti


def updatefig(*args):
    """ update the figure during the loop

        read/compute the field before and update the imshow object 'im'
        also update the title object 'ti' with the time """

    frame, nc, kt, time = args[0]
    field2d = ncpy.get_field_from_netcdf(nc, varname, kt)

    im.set_array(field2d)
    ti.set_text((title_string % varname)+time_into_string(time))

    print('\rframe=%i\b' % frame, end='')
    return im, ti


def time_into_string(time):
    """ convert the time into a suitable string for the title"""

    t = time/86400.
    day = int(t)
    t1 = (t-day)*24
    hour = int(t1)
    t2 = (t1-hour)*60
    minute = int(t2)
    string = '%id-%i:%02i' % (day, hour, minute)
    string = '%05.3f' % time
    return string


def determine_which_encoder():
    encoders = ['avconv', 'ffmpeg']
    codec = 'mpeg4'
    encoder = None
    for encod in encoders:
        if which(encod) is not None:
            encoder = encod
            break

    if encoder is None:
        print('no encoder has been found')
        raise

    else:
        print('Encoder = %s / codec = %s' % (encoder, codec))

    return encoder, codec


def browse(ncf, kz=0, varname=None):
    ncfiles = glob.glob(ncf)
    ncfiles.sort()
    for frame, nc, kt, _time in ncpy.go_through_snapshots(ncfiles):
        if varname is None:
            vnames = [v for v in nc.variables]
            varname = input('enter the variable name %s :' % vnames)
        field2d = ncpy.get_field_from_netcdf(nc, varname, kt, kz)
        if frame is 0:
            fig = plt.figure()
            plt.ion()
            im = plt.imshow(field2d, origin='lower', interpolation='nearest')
            plt.colorbar(im)
            plt.title('frame = %i' % frame)
            ax = fig.gca()
            fig.canvas.draw()
        else:
            im.set_array(field2d)
            ax.set_title('frame = %i' % frame)
            fig.canvas.draw()


if __name__ == '__main__':

    matplotlib.rcParams['font.size'] = 16

    metadata = {'title': 'CFD',
                'artist': 'Guillaume Roullet',
                'comment': "creation date: "+time.strftime("%b %d %Y")}

    my_dpi = 100
    zoom_factor = 2
    # best youtube aspect ratio is 16:9
    fig_size = np.array([1280, 720])
    fig_size = np.array([720, 1280])

    # https://docs.python.org/2/library/argparse.html#module-argparse
    parser = argparse.ArgumentParser(description='Do a movie from netfiles')
    parser.add_argument('--output', dest='video_file', type=str,
                        help='name of the video file')
    parser.add_argument('--var', dest='varname', type=str, required=True,
                        help='variable to animate')
    parser.add_argument('--cax', dest='cax', type=float, nargs=2,
                        help='two floats for the color range')
    parser.add_argument('--ncfiles', dest='ncf', type=str, required=True,
                        help='list of netcdf files, wildcard accepted')

    # save the user options
    args = parser.parse_args()
    varname = args.varname
    cax = args.cax
    if cax is None:
        cax = (None, None)
    if args.video_file is None:
        video_file = '%s.mp4' % varname
    else:
        video_file = args.video_file

    # template string for the figure title
    title_string = '%s / time = '

    # control the options
    print()
    print('varname = %s' % varname)
    print('cax = ', cax)
    print()

    ncfiles = glob.glob(args.ncf)
    ncfiles.sort()

    fig = plt.figure(figsize=fig_size/my_dpi, dpi=my_dpi)

    if len(ncfiles) is 0:
        print('no netcdf file found')

    else:
        print('netCDF files found = ')
        for frame, ncfile in enumerate(ncfiles):
            print(ncfile)
        print()

        print('Frames that will be generated')
        for frame, nc, kt, _time in ncpy.go_through_snapshots(ncfiles):
            if frame is 0:
                time_name = ncpy.get_unlimited_dimension(nc)
                # time = nc.variables[time_name][0]
                field2d = ncpy.get_field_from_netcdf(nc, varname, 0)
                im, ti = init_figure(fig, _time, field2d)
                plt.show(block=False)
            if frame % 50 is 0:
                print('frame = %i / time = %g' % (frame, _time))
        nb_frames = frame+1
        print()

        print('number of frames = %i' % nb_frames)
        print('video file = ', video_file)
        encoder, codec = determine_which_encoder()
        answer = input('Are you ok [Y/n]')

        if answer.lower() in ['', 'y']:
            anim = ani.FuncAnimation(fig, updatefig, interval=1,
                                     save_count=nb_frames,
                                     frames=ncpy.go_through_snapshots(ncfiles),
                                     repeat=False, blit=False)

        # on curie: module load swftools/0.9.2
        anim.save(video_file, writer=encoder, fps=12, codec=codec,
                  bitrate=50e3, metadata=metadata)
