import matplotlib

matplotlib.use('TkAgg')

import numpy as np
import matplotlib.pyplot as plt
from scipy import sparse
from scipy.sparse import linalg as la


#plt.ion()
import matplotlib as mpl

plt.close('all')

mpl.rcParams['lines.linewidth'] = 2
mpl.rcParams['lines.markersize'] = 12
mpl.rcParams['lines.markeredgewidth'] = 2
mpl.rcParams['xtick.labelsize'] = 14
mpl.rcParams['ytick.labelsize'] = 14
mpl.rcParams['axes.labelsize'] = 14
mpl.rcParams['axes.labelsize'] = 14


# you can tweak those parameters but in a first approach
# keep them like this
BC_type = 'Dirichlet'
Discretization = 'Centers'
regularterms = True
crossterms = False

if regularterms:
    maincoef = 4
else:
    maincoef = 0.

if crossterms:
    crosscoef = .5
    maincoef += 4*crosscoef


def set_msk(nx, ny):
    """ set the domain msk, return a mask array
    msk[j,i] == 0 : point is outside the domain
    msk[j,i] == 1 : point is inside the domain
    """
    if Discretization == 'Corners':
        h = 1./(nx-1)
    if Discretization == 'Centers':
        h = 1./nx

    msk = np.ones((ny, nx), dtype=int)

    msk[:ny//2, :nx//2] = 0
    return msk


def set_G(msk):
    """ set the array of index of interior cells
    N is the number of interior points, it is the
    number of unknowns of the Poisson problem
    G[j,i] == -1 on outside points
    G[j,i] == 0,1,2 .. N-1 on interior points
    """
    G = np.zeros_like(msk, dtype=int)-1
    idx = np.where(msk == 1)
    # i the vector of 'x' coordinates
    # j the vector of 'y' coordinates
    j, i = idx[0], idx[1]
    # N is the number of grid points such that msk != 1
    N = len(i)
    for k in range(N):
        G[j[k], i[k]] = k
    return G, i, j


def set_Laplac(G, crosscoef=0):
    """ set the sparse laplacian matrix associated
    with the domain G
    """
    crossterms = (crosscoef > 0)
    maincoef = 4+4*crosscoef
    ny, nx = np.shape(G)

    # get list of interior grid points
    idx = np.where(G > -1)
    # i the vector of 'x' coordinates
    # j the vector of 'y' coordinates
    j, i = idx[0], idx[1]
    # the Laplacian matrix is N x N
    N = len(i)

    BC = BC_type+Discretization
    # regular neighbours West, East, South and North
    listneighbours = [(-1, 0), (1, 0), (0, -1), (0, 1)]
    coefneighbours = [1]*4

    if crossterms:
        # Extended stencil with corners
        listneighbours += [(-1, -1), (-1, 1), (1, -1), (1, 1)]
        coefneighbours += [crosscoef]*4

    data = np.zeros((9*N,))
    rows = np.zeros((9*N,), dtype=int)
    cols = np.zeros((9*N,), dtype=int)
    count = 0

    for k in range(N):
        nbneighb = 0
        for who, coef in zip(listneighbours, coefneighbours):
            jj, ii = who[0], who[1]
            inside = (
                ((i[k]+ii) >= 0) and
                ((i[k]+ii) <= nx-1) and
                ((j[k]+jj) >= 0) and
                ((j[k]+jj) <= ny-1))
            # set the coefficient only if the neighbour is inside the grid
            if inside:
                l = G[j[k]+jj, i[k]+ii]
                # and is interior
                if l > -1:
                    rows[count], cols[count], data[count] = k, l, coef
                    count += 1
                    nbneighb += coef

        if BC == 'NeumannCenters':
            cff = -nbneighb
        if BC == 'NeumannCorners':
            # this case is harder to handle=>need to update the opposite neighbour
            cff = np.NaN
        if BC == 'DirichletCenters':
            cff = -(maincoef-nbneighb)-maincoef
        if BC == 'DirichletCorners':
            cff = -maincoef

        rows[count], cols[count], data[count] = k, k, cff
        count += 1

    D2 = sparse.coo_matrix(
        (data[:count], (rows[:count], cols[:count])), shape=(N, N))
    # return the sparse matrix in 'csc' format
    # the best storage for fast computations
    return D2.tocsc()/(1+2*crosscoef)


def condest(A):
    """ compute the condition number of the sparse matrix A
    A needs to normal (cf. wikipedia)
    """
    l0, w = la.eigs(A, k=1, which='SM')
    l1, w = la.eigs(A, k=1, which='LM')
    l0 = abs(l0[0])
    l1 = abs(l1[0])
    cond = l1/l0
    return cond


def Gzeros(G):
    """ return a vector of zeros that has the right number of unknowns
    for the domain G
    """
    idx = np.where(G > -1)
    return np.zeros((len(idx[0]),))


def getrhs(G):
    """ return a rhs vector that has the right number of unknowns
    for the domain G
    """
    idx = np.where(G > -1)
    n = len(idx[0])
    b = np.zeros((n,))
    # dipole configuration: two charges of opposite signs
    b[int(n*.5)] = 1.
    b[int(n*.7)] = -1.
    return b


def vec2array(x, msk, G, i, j):
    """ map the vector x onto a 2D array (for plotting)
    """
    phi = np.zeros_like(G, dtype=float)
    # i and j are N elements vectors !!!
    phi[j, i] = x
    phi[msk == 0] = np.NaN
    return phi


class Figure(object):
    def __init__(self, z2d, method):
        self.fig = plt.figure(figsize=(6, 6))
        self.ax1 = self.fig.add_subplot(1, 1, 1)
        self.im = self.ax1.imshow(z2d, extent=[0, 1, 0, 1], cmap='RdBu_r',
                                  interpolation='nearest', origin='lower')
        plt.colorbar(self.im)
        self.method = method
        self.ax1.set_title('%s / ite =%i' % (self.method, 0))
        self.ax1.set_xlabel('X')
        self.ax1.set_ylabel('Y')
        self.fig.show()
        self.fig.canvas.draw()

    def update(self, z2d, kt):
        self.im.set_array(z2d)
        colorscheme = 'auto'
        if colorscheme == 'imposed':
            vmin, vmax = cax
        elif colorscheme == 'auto':
            vmin, vmax = np.nanmin(z2d), np.nanmax(z2d)
        else:
            raise ValueError('colorscheme should be "imposed" or "auto"')
        self.ax1.set_title('%s / ite =%i' % (self.method, kt))
        self.im.set_clim(vmin=vmin, vmax=vmax)
        self.fig.canvas.draw()


def jacobi(A, b, x0=None, omega=1., tol=1e-6, maxiter=200, callback=None):
    """ Jacobi solver

    A is split into A = D + R, with D the diagonal

    the iteration consists of x_{k+1} = D^{-1}(b-R*x_k)
    """
    ite = 0
    if x0 is None:
        x0 = b*0
    res = A*x0-b
    err = np.linalg.norm(res)
    tol = tol*np.linalg.norm(b)
    d = A.diagonal()
    R = A-sparse.diags(d)
    x = x0.copy()
    while (err > tol) and (ite < maxiter):
        x = (1-omega)*x+omega*(b-R*x)/d
        res = A*x-b
        err = np.linalg.norm(res)
        ite += 1
        if callback != None:
            callback(x)
    if ite < maxiter:
        info = 0
    else:
        info = 1
    return x, info


def solveiter(A, b, method, grid, plot=False):
    """ solve A*x=b using the iterative 'method'
    return the error at each iteration
    """
    msk, G, i, j = grid

    def checkerror(xk):
        global ite
        ite += 1
        d = np.linalg.norm(xk-xexact)
        err.append(d)
        if plot:
            fig.update(vec2array(xk-xexact, msk, G, i, j), ite)

    err = []
    if method == 'cg':
        meth = la.cg
    elif method == 'cgs':
        meth = la.iterative.cgs
    elif method == 'qmr':
        meth = la.iterative.qmr
    elif method == 'jacobi':
        meth = jacobi

    xexact = la.spsolve(A, b)
    global ite
    ite = 0
    if plot:
        fig = Figure(vec2array(xexact, msk, G, i, j), method)
    x, info = meth(A, b, tol=1e-10, callback=checkerror)
    if info != 0:
        print('pb with convergence / method=%s /  info=%i' % (method, info))
    return err


def conv_vs_method(plot=False):
    """ convergence vs the iterative method
    """
    plt.figure(1)

    nx = 50
    ny = nx
    msk = set_msk(nx, ny)
    G, i, j = set_G(msk)
    D2 = set_Laplac(G)
    b = getrhs(G)
    grid = [msk, G, i, j]
    for m in ['cg', 'cgs', 'qmr', 'jacobi']:
        err = solveiter(D2, b, m, grid, plot=plot)
        plt.figure(1)
        plt.semilogy(err, label=m)
    plt.xlabel('iteration')
    plt.ylabel('||x-x_e||')
    plt.legend()
    plt.grid()


def conv_vs_resolution():
    """ convergence vs resolution
    """
    plt.figure(2)

    print('-'*70)
    print('Condition number for the Laplacian operator')
    for nx in [2**k for k in range(3, 8)]:
        ny = nx
        msk = set_msk(nx, ny)
        G, i, j = set_G(msk)
        D2 = set_Laplac(G)
        cond = condest(D2)

        print('nx=%3i / condition number : %.1f' % (nx, cond))

        b = Gzeros(G)+1
        b = b/(nx*nx)
        grid = [msk, G, i, j]
        err = solveiter(D2, b, 'cg', grid)
        plt.semilogy(err, label='n=%i' % nx)
    plt.xlabel('iteration')
    plt.ylabel('||x-x_e||')
    plt.legend()
    plt.grid()


def condnumber_helmholtz():
    """ condition number of the Helmoltz operator vs alpha
    """
    nx = 50
    ny = nx
    msk = set_msk(nx, ny)
    G, i, j = set_G(msk)
    D2 = set_Laplac(G)
    b = getrhs(G)

    dx = 1./nx
    N = D2.shape[0]
    I = sparse.identity(N)
    print('-'*70)
    print('Condition number for the Helmholtz operator')
    for alpha in [0., 0.1, 0.5, 1, 2, 10, 100, 1000, 1e4]:
        H = I-alpha*D2
        cond = condest(H)
        print('alpha=%8.2f / condition number : %.1f' % (alpha, cond))

    cond = condest(D2)
    print('alpha=     Inf / condition number : %.1f' % (cond))

    plt.figure()
    alpha = 10
    H = I-alpha*D2
    xexact = la.spsolve(H, b)
    phi = vec2array(xexact, msk, G, i, j)
    plt.imshow(phi, interpolation='nearest', origin='xy')
    plt.title(r'solution of $(1 -\alpha\, dx ^ 2\, \nabla ^ 2)\phi=b$, '
              + r'$\alpha$= % 3.1f' % alpha, fontsize=18)
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.colorbar()


if __name__ == "__main__":

    conv_vs_method(plot=False)

    # conv_vs_resolution()

    # condnumber_helmholtz()

    plt.show()
